import 'dart:async';
import 'dart:convert';

import 'package:dio/dio.dart';

///使用Dio内置FormData时，服务端无法解析韩文参数值，暂时使用CustomFormData
class CustomFormData extends FormData {
  static const String _BOUNDARY_PRE_TAG = '--dio-boundary-';
  static const _BOUNDARY_LENGTH = _BOUNDARY_PRE_TAG.length + 10;

  final _newlineRegExp = RegExp(r'\r\n|\r|\n');

  bool get isFinalized => _mIsFinalized;
  bool _mIsFinalized = false;

  int get length {
    var length = 0;
    fields.forEach((entry) {
      length += '--'.length +
          _BOUNDARY_LENGTH +
          '\r\n'.length +
          utf8.encode(_headerForField(entry.key, entry.value)).length +
          utf8.encode(entry.value).length +
          '\r\n'.length;
    });

    for (var file in files) {
      length += '--'.length +
          _BOUNDARY_LENGTH +
          '\r\n'.length +
          utf8.encode(_headerForFile(file)).length +
          file.value.length +
          '\r\n'.length;
    }

    return length + '--'.length + _BOUNDARY_LENGTH + '--\r\n'.length;
  }

  CustomFormData(): super();

  CustomFormData.fromMap(Map map) : super() {
    encodeMap(
      map,
      (key, value) {
        if (value == null) return null;
        if (value is MultipartFile) {
          files.add(MapEntry(key, value));
        } else {
          fields.add(MapEntry(key, value.toString()));
        }
        return null;
      },
      encode: false,
    );
  }

  String _browserEncode(String value) {
    return value.replaceAll(_newlineRegExp, '%0D%0A').replaceAll('"', '%22');
  }

  String _headerForField(String name, String value) {
    var header =
        'content-disposition: form-data; name="${_browserEncode(name)}"';
    return '$header\r\n\r\n';
  }

  String _headerForFile(MapEntry<String, MultipartFile> entry) {
    var file = entry.value;
    var header =
        'content-disposition: form-data; name="${_browserEncode(entry.key)}"';
    if (file.filename != null) {
      header = '$header; filename="${_browserEncode(file.filename??'')}"';
    }
    header = '$header\r\n'
        'content-type: ${file.contentType}';
    return '$header\r\n\r\n';
  }

  @override
  Stream<List<int>> finalize() {
    if (isFinalized) {
      throw StateError("Can't finalize a finalized MultipartFile.");
    }
    _mIsFinalized = true;
    var controller = StreamController<List<int>>(sync: false);
    void writeAscii(String string) {
      controller.add(utf8.encode(string));
    }

    void writeUtf8(String string) => controller.add(utf8.encode(string));
    void writeLine() => controller.add([13, 10]); // \r\n

    fields.forEach((entry) {
      writeAscii('--$boundary\r\n');
      writeAscii(_headerForField(entry.key, entry.value));
      writeUtf8(entry.value);
      writeLine();
    });

    Future.forEach<MapEntry<String, MultipartFile>>(files, (file) {
      writeAscii('--$boundary\r\n');
      writeAscii(_headerForFile(file));
      return writeStreamToSink(file.value.finalize(), controller)
          .then((_) => writeLine());
    }).then((_) {
      writeAscii('--$boundary--\r\n');
      controller.close();
    });
    return controller.stream;
  }
}

Future writeStreamToSink(Stream stream, EventSink sink) {
  var completer = Completer();
  stream.listen(sink.add,
      onError: sink.addError, onDone: () => completer.complete());
  return completer.future;
}

typedef DioEncodeHandler = Function(String key, Object value);

String encodeMap(data, DioEncodeHandler handler, {bool encode = true}) {
  var urlData = StringBuffer('');
  var first = true;
  var leftBracket = encode ? '%5B' : '[';
  var rightBracket = encode ? '%5D' : ']';
  var encodeComponent = encode ? Uri.encodeQueryComponent : (e) => e;
  void urlEncode(dynamic sub, String path) {
    if (sub is List) {
      for (var i = 0; i < sub.length; i++) {
        urlEncode(sub[i],
            '$path$leftBracket${(sub[i] is Map || sub[i] is List) ? i : ''}$rightBracket');
      }
    } else if (sub is Map) {
      sub.forEach((k, v) {
        if (path == '') {
          urlEncode(v, '${encodeComponent(k)}');
        } else {
          urlEncode(v, '$path$leftBracket${encodeComponent(k)}$rightBracket');
        }
      });
    } else {
      var str = handler(path, sub);
      var isNotEmpty = str != null && str.trim().isNotEmpty;
      if (!first && isNotEmpty) {
        urlData.write('&');
      }
      first = false;
      if (isNotEmpty) {
        urlData.write(str);
      }
    }
  }

  urlEncode(data, '');
  return urlData.toString();
}
